import 'dart:developer';

import 'package:hydrated_bloc/hydrated_bloc.dart';

import '../../loadableState/loading_error.dart';
import '../../repository/repository.dart';
import 'loadable_hydrated_bloc_event.dart';
import '../../loadableState/loadable_state.dart';

abstract class LoadableHydratedBloc<
  TEvent extends LoadableHydratedBlocEvent<TState>,
  TState,
  TRepository extends Repository<TState>
> extends HydratedBloc<
    LoadableHydratedBlocEvent<TState>,
    LoadableState<TState>
> {
  late TRepository _repository;
  LoadableHydratedBloc() : super(const LoadableState()) {
    on<Emit<TState>>((event, emit) => emit(LoadableState(isLoading: event.loading, data: event.state(innerState ?? fromNothing()), reFetch: retry)));
    on<RefetchStarted<TState>>((event, emit) => emit(LoadableState(isLoading: true, data: innerState)));
    on<ClearState<TState>>((event, emit) => emit(const LoadableState()));
    on<Error<TState>>((event, emit) => emit(LoadableState(isLoading: false, data: innerState, reFetch: retry, error: event.error)));

    _repository = repository();
    fetch();
  }

  TState? get innerState => state.data;
  TRepository get repo => _repository;

  void retry() {
    log('Fetch retry triggered for ${TState.toString()}');
    add(RefetchStarted<TState>());
    fetch();
  }

  void fetch() {
    log('Fetching data for ${TState.toString()}');
    gatherData().catchError(
      (e) {
        log('Error while fetching ${TState.toString()}: ${e.toString()}');
        add(Error(LoadingError(
          message: e.toString(),
          allowRetry: true,
        )));
      }
    ).then((value) {
      log('Fetch for ${TState.toString()} completed!');
    });
  }

  @override
  fromJson(Map<String, dynamic> json) => LoadableState(isLoading: true, data: fromStorage(json));
  @override
  Map<String, dynamic>? toJson(LoadableState<TState> state) => state.data == null ? {} : state.data.toJson();

  Future<void> gatherData();
  TRepository repository();

  TState fromNothing();
  TState fromStorage(Map<String, dynamic> json);
  Map<String, dynamic>? toStorage(TState state);
}
